from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordUISerializerBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_ui_serializer"
    class_config = "record-ui-serializer-class"
    template = "record-ui-serializer"

    def finish(self, **extra_kwargs):

        ui_record_schema_class = (
            self.current_model.get("ui", {})
            .get("marshmallow", {})
            .get("schema-class", self.current_model.record_schema_class)
        )

        return super().finish(
            ui_record_schema_class=ui_record_schema_class, **extra_kwargs
        )



class SchemaPathValidator:
    valid = False

    def get(self, path_el):
        raise NotImplementedError()

    @property
    def real(self):
        return self


class Invalid(SchemaPathValidator):
    def get(self, path_el):
        return self


invalid = Invalid()


class PrimitiveValidator(SchemaPathValidator):
    valid = True

    def get(self, path_el):
        return invalid


primitive = PrimitiveValidator()


class DictValidator(SchemaPathValidator):
    valid = True

    def __init__(self, dict=None, primitives=None):
        self.dict = dict or {}
        if primitives:
            for p in primitives.split(","):
                p = p.strip()
                if not p:
                    continue
                self.dict[p] = primitive

    def get(self, path_el):
        if path_el in self.dict:
            return self.dict[path_el]
        return invalid


class AnyKeyDictValidator(SchemaPathValidator):
    valid = True

    def __init__(self, _next):
        self._next = _next

    def get(self, path_el):
        return self._next


class ArrayValidator(AnyKeyDictValidator):
    pass


class Ref(SchemaPathValidator):
    refs = {}
    valid = True

    def __init__(self, element_type, refstr):
        self.element_type = element_type
        self.refstr = refstr

    def get(self, path_el):
        return self.refs[self.refstr].get(path_el)

    @property
    def real(self):
        return self.refs[self.refstr]

    def __str__(self):
        return self.element_type

    def __repr__(self):
        return str(self)


Ref.refs["type"] = DictValidator(
    dict={
        "properties": Ref("properties", "properties"),
        "patternProperties": Ref("patternProperties", "properties"),
        "additionalProperties": Ref("additionalProperties", "additionalProperties"),
        "propertyNames": Ref("propertyNames", "propertyNames"),
        "items": Ref("items", "type"),
        "prefixItems": Ref("prefixItems", "type"),
        "contains": Ref("contains", "type"),
        "type": Ref("type", "type_value"),
        "enum": Ref("schemaConstraint", "arr_constraint"),  # enums and consts
        "const": Ref("schemaConstraint", "constraint"),
        "required": Ref("schemaConstraint", "constraint"),  # object
        "minProperties": Ref("schemaConstraint", "constraint"),
        "maxProperties": Ref("schemaConstraint", "constraint"),
        "minItems": Ref("schemaConstraint", "constraint"),  # array
        "maxItems": Ref("schemaConstraint", "constraint"),
        "uniqueItems": Ref("schemaConstraint", "constraint"),
        "minContains": Ref("schemaConstraint", "constraint"),
        "maxContains": Ref("schemaConstraint", "constraint"),
        "minLength": Ref("schemaConstraint", "constraint"),  # string
        "maxLength": Ref("schemaConstraint", "constraint"),
        "pattern": Ref("schemaConstraint", "constraint"),
        "format": Ref("schemaConstraint", "constraint"),
        "minimum": Ref("schemaConstraint", "constraint"),  # numbers
        "exclusiveMinimum": Ref("schemaConstraint", "constraint"),
        "maximum": Ref("schemaConstraint", "constraint"),
        "exclusiveMaximum": Ref("schemaConstraint", "constraint"),
        "multipleOf": Ref("schemaConstraint", "constraint"),
    }
)

Ref.refs["additionalProperties"] = DictValidator(primitives="type")

Ref.refs["propertyNames"] = DictValidator(primitives="pattern")

Ref.refs["properties"] = AnyKeyDictValidator(Ref("property", "type"))

Ref.refs["type_value"] = PrimitiveValidator()

Ref.refs["constraint"] = PrimitiveValidator()

Ref.refs["arr_constraint"] = ArrayValidator(Ref("arr_constraint_item", "constraint"))

schema_paths = DictValidator(
    {
        "$vocabulary": AnyKeyDictValidator(primitive),
        "properties": Ref("properties", "properties"),
    },
    primitives="$schema,$id,type",
)

model_paths = schema_paths



from pathlib import Path

from oarepo_model_builder.loaders import json_loader
from oarepo_model_builder.schema import ModelSchema

DUMMY_PATH = "/tmp/path.json"  # NOSONAR checking as it is a virtual path


def test_loading_from_string():
    schema = ModelSchema(DUMMY_PATH, {})
    assert schema.schema == {"settings": {}}


def test_loading_from_empty_file():
    schema = ModelSchema(
        Path(__file__).parent.joinpath("data/empty.json"), loaders={"json": json_loader}
    )
    assert schema.schema == {"settings": {}}


def test_loading_included_resource():
    schema = ModelSchema(
        DUMMY_PATH,
        {"a": {"use": "test1"}},
        {"test1": lambda schema: {"included": "test1"}},
    )
    assert schema.schema == {
        "settings": {},
        "a": {"included": "test1"},
    }


def test_loading_included_resource_root():
    schema = ModelSchema(
        DUMMY_PATH,
        {"use": "test1"},
        {"test1": lambda schema: {"included": "test1"}},
    )
    assert schema.schema == {
        "settings": {},
        "included": "test1",
    }


def test_loading_jsonpath_resource():
    schema = ModelSchema(
        DUMMY_PATH,
        {"use": "test1#/test/a"},
        {"test1": lambda schema: {"test": {"a": {"included": "test1"}}}},
    )
    assert schema.schema == {
        "settings": {},
        "included": "test1",
    }


def test_loading_current():
    schema = ModelSchema(DUMMY_PATH, {"b": {"use": "#/a"}, "a": {"a": True}})
    assert schema.schema == {
        "settings": {},
        "b": {"a": True},
        "a": {"a": True},
    }


def test_loading_current_by_id():
    schema = ModelSchema(
        DUMMY_PATH,
        {"b": {"use": "#id"}, "a": {"$id": "id", "a": True}},
    )
    assert schema.schema == {
        "settings": {},
        "b": {"a": True},
        "a": {"$id": "id", "a": True},
    }


def test_loading_external_by_id():
    schema = ModelSchema(
        DUMMY_PATH,
        {
            "b": {"use": "aa#id"},
        },
        {"aa": lambda schema: {"a": {"$id": "id", "a": True}}},
    )
    assert schema.schema == {
        "settings": {},
        "b": {"a": True},
    }



from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.model_preprocessors import ModelPreprocessor
from oarepo_model_builder.schema import ModelSchema


def test_empty_builder():
    builder = ModelBuilder(output_builders=[], outputs=[], property_preprocessors=[])
    builder.skip_schema_validation = True
    outputs = builder.build(ModelSchema("", {"a": 1}), "/tmp/test")
    assert outputs == {}


def test_transformer():
    called = [False]

    class SampleModelPreprocessor(ModelPreprocessor):
        def transform(self, schema, settings):
            schema.set("test", 1)
            called[0] = True

    builder = ModelBuilder(
        output_builders=[],
        outputs=[],
        property_preprocessors=[],
        model_preprocessors=[SampleModelPreprocessor],
    )
    builder.skip_schema_validation = True
    schema = ModelSchema("", {"a": 2})
    builder.build(schema, "/tmp/test")

    assert called[0]



import copy
import importlib
from pathlib import Path
from typing import Dict, List, Type, Union

from .builders import OutputBuilder, OutputBuilderComponent
from .fs import AbstractFileSystem, FileSystem
from .model_preprocessors import ModelPreprocessor
from .outputs import OutputBase
from .property_preprocessors import PropertyPreprocessor
from .schema import ModelSchema
from .utils.cst import ConflictResolver
from .validation import validate_model


class ModelBuilder:
    """
    Processes a model file and generates/updates sources for the model
    """

    model_preprocessor_classes: List[Type[ModelPreprocessor]]
    """
    Model preprocessor classes that are called after schema is loaded and before it is processed
    """

    output_classes: List[Type[OutputBase]]
    """
    Mapping between output type and its handler class
    """

    filtered_output_classes: Dict[str, Type[OutputBase]]
    """
    Filtered output classes by settings.plugins.disabled|enabled
    """

    output_builder_classes: List[Type[OutputBuilder]]
    """
    A list of extension classes to be used in build. 
    """

    property_preprocessor_classes: List[Type[PropertyPreprocessor]]
    """
    Processor classes (called before and after file builder is called)
    """

    output_builders: List[OutputBuilder]
    """
    A list of output_builders. Each extension is responsible for generating one or more files
    """

    output_builder_components: Dict[str, List[OutputBuilderComponent]]
    """
    A list of output builder components for an output builder
    """

    outputs: Dict[Path, OutputBase]
    """
    Mapping between concrete output (file path relative to output dir) and instance of builder class
    """

    property_preprocessors: List[PropertyPreprocessor]
    """
    Current instances of processor classes.
    """

    filesystem: AbstractFileSystem

    overwrite: bool
    """
    If true, overwrite already existing files. If false, perform merge
    """

    conflict_resolver: ConflictResolver
    """
    Resolver for conflicts
    """

    def __init__(
        self,
        outputs: List[Type[OutputBase]] = (),
        output_builders: List[Type[OutputBuilder]] = (),
        property_preprocessors: List[Type[PropertyPreprocessor]] = (),
        model_preprocessors: List[Type[ModelPreprocessor]] = (),
        output_builder_components: Dict[str, List[Type[OutputBuilderComponent]]] = None,
        filesystem=FileSystem(),
        conflict_resolver: ConflictResolver = None,
        overwrite=False,
    ):
        """
        Initializes the builder

        :param output_builders:          A list of extension classes to use in builds
        :param outputs:     List of file builder classes that generate files
        :param property_preprocessors: List of output type processor classes
        """
        self.output_builder_classes = [*output_builders]
        for o in outputs:
            assert o.TYPE, f"output_type not set up on class {o}"
        self.output_classes = [*(outputs or [])]
        self.outputs = {}
        self.property_preprocessor_classes = [*(property_preprocessors or [])]
        self.model_preprocessor_classes = [*(model_preprocessors or [])]
        self.filtered_output_classes = {o.TYPE: o for o in self.output_classes}
        if output_builder_components:
            self.output_builder_components = {
                builder_type: [x() for x in components]
                for builder_type, components in output_builder_components.items()
            }
        else:
            self.output_builder_components = {}
        self.filesystem = filesystem
        self.skip_schema_validation = False  # set to True in some tests
        self.conflict_resolver = conflict_resolver
        self.overwrite = overwrite

    def get_output(self, output_type: str, path: Union[str, Path]):
        """
        Given a path, instantiate file builder on the path with the given output type
        and return it. If the builder on the path has already been requested, return
        the same instance of the builder.

        :param output_type: @see FileBuilder.output_type
        :param path: relative path to output_dir, set in build()
        :return:    instance of FileBuilder for the path
        """
        if not isinstance(path, Path):
            path = Path(path)
        path = self.output_dir.joinpath(path)

        output = self.outputs.get(path, None)
        if output:
            assert output_type == self.outputs[path].TYPE
        else:
            output = self.filtered_output_classes[output_type](self, path)
            output.begin()
            self.outputs[path] = output
        return output

    def get_output_builder_components(self, output_builder_type):
        return self.output_builder_components.get(output_builder_type, ())

    # main entry point
    def build(
        self,
        model: ModelSchema,
        output_dir: Union[str, Path],
    ):
        """
        compile the schema to output directory

        :param model:      the model schema
        :param output_dir:  output directory where to put generated files
        :return:            the outputs (self.outputs)
        """

        # deep copy so that the passed model will not be influenced by preprocessors etc.
        model = copy.deepcopy(model)
        if self.overwrite:
            if not hasattr(self.filesystem, "overwrite"):
                raise AttributeError(
                    f"Filesystem of type {type(self.filesystem)} does not support overwrite"
                )
            self.filesystem.overwrite = True

        self.set_schema(model)
        self.filtered_output_classes = {
            o.TYPE: o for o in self._filter_classes(self.output_classes, "output")
        }
        self.output_dir = Path(output_dir).absolute()  # noqa
        self.outputs = {}

        self._validate_model(model)

        self._run_model_preprocessors(model)
        self._validate_model(model)

        # noinspection PyTypeChecker
        property_preprocessors: List[PropertyPreprocessor] = [
            e(self)
            for e in self._filter_classes(
                self.property_preprocessor_classes, "property"
            )
        ]

        self._run_output_builders(model, property_preprocessors)

        self._save_outputs()

        return self.outputs

    def _run_output_builders(self, model, property_preprocessors):
        output_builder_class: Type[OutputBuilder]
        for output_builder_class in self._filter_classes(
            self.output_builder_classes, "builder"
        ):
            output_builder = output_builder_class(
                builder=self, property_preprocessors=property_preprocessors
            )
            output_builder.build(model)

    def _run_model_preprocessors(self, model):
        for model_preprocessor in self._filter_classes(
            self.model_preprocessor_classes, "model"
        ):
            model_preprocessor(self).transform(model, model.settings)

    def _validate_model(self, model):
        if not self.skip_schema_validation:
            validate_model(model)

    def _save_outputs(self):
        for output in sorted(self.outputs.values(), key=lambda x: x.path):
            output.finish()
            if output.executable:
                self.filesystem.make_executable(output.path)

    def set_schema(self, schema):
        self.schema = schema
        self.settings = schema.settings

    # private methods

    def _filter_classes(self, classes: List[Type[object]], plugin_type):
        if (
            "plugins" not in self.schema.current_model
            or plugin_type not in self.schema.current_model.plugins
        ):
            return classes
        plugin_config = self.schema.current_model.plugins[plugin_type]

        disabled = plugin_config.get("disable", [])
        enabled = plugin_config.get("enable", [])
        included = plugin_config.get("include", [])

        if included:
            enabled = [*enabled]  # will be adding inclusions so make a copy
            classes = [*classes]
            for incl in included:
                package_name, class_name = incl.split(":")
                class_type = getattr(importlib.import_module(package_name), class_name)
                classes.append(class_type)
                if enabled and class_type.TYPE not in enabled:
                    enabled.append(class_type.TYPE)

        if disabled == "__all__":
            ret = []
        else:
            ret = [c for c in classes if c.TYPE not in disabled]

        if enabled:
            ret.extend([c for c in classes if c.TYPE in enabled])
        return ret



import json
import os
import re
import sys
from pathlib import Path

from oarepo_model_builder.entrypoints import create_builder_from_entrypoints, load_model
from oarepo_model_builder.fs import InMemoryFileSystem
from tests.utils import assert_python_equals

OAREPO_USE = "use"


def test_include_invenio():
    schema = load_model(
        "test.yaml",
        "test",
        model_content={
            "version": "1.0.0",
            "model": {OAREPO_USE: "invenio", "properties": {"a": {"type": "keyword"}}},
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    data = builder.filesystem.open(
        os.path.join("test", "services", "records", "schema.py")
    ).read()
    print(data)
    assert re.sub(r"\s", "", data) == re.sub(
        r"\s",
        "",
        """
from invenio_records_resources.services.records.schema import BaseRecordSchema as InvenioBaseRecordSchema
from marshmallow import ValidationError
from marshmallow import validate as ma_validate
import marshmallow as ma
from marshmallow import fields as ma_fields
from marshmallow_utils import fields as mu_fields
from marshmallow_utils import schemas as mu_schemas
from oarepo_runtime.ui import marshmallow as l10n
 from oarepo_runtime.validation import validate_date

class TestSchema(InvenioBaseRecordSchema):
    \"""TestSchema schema.\"""
    a = ma_fields.String()
    created = ma_fields.String(validate=[validate_date('%Y-%m-%d')], dump_only=True)
    updated = ma_fields.String(validate=[validate_date('%Y-%m-%d')], dump_only=True)
    """,
    )

    data = builder.filesystem.read(
        os.path.join("test", "records", "mappings", "os-v2", "test", "test-1.0.0.json")
    )
    data = json.loads(data)
    assert data == {
        "mappings": {
            "properties": {
                "$schema": {"type": "keyword"},
                "a": {"type": "keyword"},
                "created": {"type": "date"},
                "id": {"type": "keyword"},
                "updated": {"type": "date"},
            },
        }
    }

    data = builder.filesystem.read("setup.cfg")
    assert f"version = 1.0.0" in data


def test_generate_multiple_times():
    schema = load_model(
        "test.yaml",
        "test",
        model_content={
            "model": {OAREPO_USE: "invenio", "properties": {"a": {"type": "keyword"}}},
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")
    snapshot_1 = filesystem.snapshot()

    # need to reload the schema because of caches ...
    schema = load_model(
        "test.yaml",
        "test",
        model_content={
            "model": {OAREPO_USE: "invenio", "properties": {"a": {"type": "keyword"}}},
        },
        isort=False,
        black=False,
    )
    builder.build(schema, "")
    snapshot_2 = filesystem.snapshot()

    assert snapshot_1.keys() == snapshot_2.keys()

    for k in snapshot_1:
        first_val = snapshot_1[k]
        second_val = snapshot_2[k]
        if first_val != second_val:
            print("ASSERT FAILED", k, file=sys.stderr)
            print("====================", file=sys.stderr)
            print(first_val, file=sys.stderr)
            print("====================", file=sys.stderr)
            print(second_val, file=sys.stderr)
            print("====================", file=sys.stderr)
            assert first_val == second_val


def test_incremental_builder():
    schema = load_model(
        "test.yaml",
        "test",
        model_content={
            "model": {OAREPO_USE: "invenio", "properties": {"a": {"type": "keyword"}}},
        },
        isort=False,
        black=False,
    )

    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")

    schema = load_model(
        "test.yaml",
        "test",
        model_content={
            "model": {OAREPO_USE: "invenio", "properties": {"a": {"type": "keyword"}}},
        },
        isort=False,
        black=False,
    )

    builder.build(schema, "")
    snapshot_1 = filesystem.snapshot()

    snapshot_1.pop(
        Path.cwd() / "data/sample_data.yaml"
    )  # these are always regenerated and random, so do not check them

    schema = load_model(
        "test.yaml",
        "test",
        model_content={
            "model": {OAREPO_USE: "invenio", "properties": {"a": {"type": "keyword"}}},
        },
        isort=False,
        black=False,
    )
    filesystem = InMemoryFileSystem()
    builder = create_builder_from_entrypoints(filesystem=filesystem)

    builder.build(schema, "")
    snapshot_2 = filesystem.snapshot()

    ret = snapshot_2.pop(
        Path.cwd() / "data/sample_data.yaml"
    )  # these are always regenerated and random, so do not check them

    assert set(snapshot_1.keys()) == set(snapshot_2.keys())

    for k, iteration_result in snapshot_1.items():
        expected_result = snapshot_2[k]
        # normally handled by black
        assert_python_equals(
            iteration_result.replace(",'_id'", ",\n'_id'"), expected_result, f"File {k}"
        )



import copy
import pathlib
from pathlib import Path
from typing import Callable, Dict, List, Union

import munch
import yaml
from jsonpointer import resolve_pointer
from yaml import SafeDumper

from .exceptions import IncludedFileNotFoundException
from .utils.deepmerge import deepmerge
from .utils.hyphen_munch import HyphenMunch


class Key(str):
    def __new__(cls, value, *args, source=None, **kwargs):
        ret = super().__new__(cls, value)
        ret.sources = {source} if source else {}
        return ret

    @staticmethod
    def annotate_keys_with_source(data, source):
        if isinstance(data, dict):
            return {
                Key(k, source=source): Key.annotate_keys_with_source(v, source)
                for k, v in data.items()
            }
        elif isinstance(data, (tuple, list)):
            return [Key.annotate_keys_with_source(k, source) for k in data]
        else:
            return data

    @staticmethod
    def get_sources(data):
        if isinstance(data, Key):
            return data.sources
        return []


def key_representer(dumper, data):
    return dumper.represent_str(str(data))


yaml.add_representer(Key, key_representer)
yaml.add_multi_representer(Key, key_representer)
SafeDumper.add_representer(Key, key_representer)
SafeDumper.add_multi_representer(Key, key_representer)


class ModelSchema:
    USE_KEYWORD = "use"
    REF_KEYWORD = "$ref"

    def __init__(
        self,
        file_path,
        content=None,
        included_models: Dict[str, Callable] = None,
        merged_models: List[Union[str, Path]] = None,
        loaders=None,
        model_field="model",
    ):
        """
        Creates and parses model schema

        :param file_paths: list of paths on the filesystem to the model schema files. The content of these files will be merged before the processing.
        :param content:   if set, use this content, otherwise load the file_path
        :param included_models: a dictionary of file_id to callable that returns included json.
                The callable expects a single parameter, an instance of this schema
        """

        self.file_path = file_path
        self.included_schemas = included_models or {}
        self.loaders = loaders

        if content is not None:
            self.schema = content
        else:
            self.schema = copy.deepcopy(self._load(file_path))
            for fp in merged_models or []:
                self.schema = deepmerge(self.schema, copy.deepcopy(self._load(fp)))

        self._resolve_references(self.schema, [])

        self._resolve_shortcuts(self.schema)

        # any star keys should be kept
        use_star_keys(self.schema)

        self.schema.setdefault("settings", {})
        self.schema = munch.munchify(self.schema, factory=HyphenMunch)

        self.model_field = model_field

    def debug_print(self):
        def _print(data, prefix):
            if isinstance(data, dict):
                print()
                for k, v in sorted(data.items()):
                    print(f"{prefix}{k}{Key.get_sources(k)}:", end="")
                    _print(v, prefix + "  ")
            elif isinstance(data, (list, tuple)):
                print()
                for v in sorted(data):
                    _print(v, prefix + "-  ")
            else:
                print(f"{prefix}{data}")

        _print(self.schema, "")
        print()

    def get(self, key):
        return self.schema.get(key, None)

    def set(self, key, value):
        self.schema[key] = value

    @property
    def settings(self):
        return self.schema.settings

    @property
    def current_model(self):
        return self.schema.get(self.model_field, {})

    def merge(self, another):
        self.schema = munch.munchify(
            deepmerge(another, self.schema, []), factory=HyphenMunch
        )

    def _load(self, file_path, content=None):
        """
        Loads a json/json5 file on the path

        :param file_path: file path on filesystem
        :return: parsed json
        """
        if file_path in self.included_schemas:
            loaded = self._fetch_included(file_path)
        else:
            extension = pathlib.Path(file_path).suffix.lower()[1:]
            if extension not in self.loaders:
                raise RuntimeError(
                    f"Can not load {file_path} - no loader has been found for extension {extension} "
                    f"in entry point group oarepo_model_builder.loaders"
                )
            loaded = self.loaders[extension](file_path, self, content=content)
        return Key.annotate_keys_with_source(loaded, file_path)

    def _fetch_included(self, file_path):
        included = self.included_schemas[file_path]
        if callable(included):
            return included(self)
        extension = pathlib.Path(included).suffix.lower()[1:]
        if extension not in self.loaders:
            raise RuntimeError(
                f"Can not load {included} - no loader has been found for extension {extension} "
                f"in entry point group oarepo_model_builder.loaders"
            )
        return self.loaders[extension](included, self)

    def _load_included_file(self, location, source_locations=None):
        """
        Resolve and load an included file. Internal method called when loading schema.
        If the included file contains a json pointer,
        return only the part identified by the json pointer.

        :param file_id: the id of the included file, might contain #xpointer
        :return:    loaded json
        """
        if "#" in location:
            file_id, json_pointer_or_id = location.rsplit("#", 1)
        else:
            file_id = location
            json_pointer_or_id = None

        if not file_id or file_id == ".":
            ret = self.schema
        else:
            file_path = self._resolve_file_path(file_id, source_locations)
            if file_path:
                # relative include
                ret = self._load(file_path)
            else:
                if file_id not in self.included_schemas:
                    raise IncludedFileNotFoundException(
                        f"Included file {file_id} not found in includes"
                    )

                ret = self.included_schemas[file_id](self)

        if json_pointer_or_id:
            if json_pointer_or_id.startswith("/"):
                ret = resolve_pointer(ret, json_pointer_or_id)
            else:
                ret = resolve_id(ret, json_pointer_or_id)
                if not ret:
                    raise IncludedFileNotFoundException(
                        f"Element with id {json_pointer_or_id} not found in {file_id}"
                    )

        ret = copy.deepcopy(ret)
        if ret is None:
            ret = {}
        ret.pop("$id", None)
        return ret

    def _resolve_file_path(self, file_id, source_locations):
        for location in source_locations:
            pth = Path(location).parent / file_id
            if pth.exists():
                return pth
        pth = self.abs_path.parent / file_id
        if pth.exists():
            return pth
        return None

    def _resolve_shortcuts(self, element):
        if isinstance(element, dict):
            for v in element.values():
                self._resolve_shortcuts(v)
        elif isinstance(element, list):
            for v in element:
                self._resolve_shortcuts(v)

    def _resolve_references(self, element, stack):
        if isinstance(element, dict):
            if self.USE_KEYWORD in element or self.REF_KEYWORD in element:
                for key in element:
                    if key in (self.USE_KEYWORD, self.REF_KEYWORD):
                        break
                else:
                    raise  # just for making pycharm happy
                included_name = element[key]

                # if it is a dictionary, then probably it is a name of a property,
                # so keep it
                if isinstance(included_name, dict):
                    return self._resolve_references(element, stack)

                element.pop(self.USE_KEYWORD, None)
                element.pop(self.REF_KEYWORD, None)

                if not isinstance(included_name, list):
                    included_name = [included_name]
                for name in included_name:
                    if not name:
                        raise IncludedFileNotFoundException(
                            f"No file for use at path {'/'.join(stack)}"
                        )
                    included_data = self._load_included_file(
                        name, source_locations=Key.get_sources(key)
                    )
                    deepmerge(element, included_data, [], listmerge="keep")
                return self._resolve_references(element, stack)
            for k, v in element.items():
                self._resolve_references(v, stack + [k])
        elif isinstance(element, list):
            for v in element:
                self._resolve_references(v, stack)

    @property
    def abs_path(self):
        return Path(self.file_path).absolute()


def resolve_id(json, element_id):
    if isinstance(json, dict):
        if "$id" in json and json["$id"] == element_id:
            return json
        continue_with = json.values()
    elif isinstance(json, (tuple, list)):
        continue_with = json
    else:
        return None
    for k in continue_with:
        ret = resolve_id(k, element_id)
        if ret is not None:
            return ret


def remove_star_keys(schema):
    if isinstance(schema, dict):
        for k, v in list(schema.items()):
            if k.startswith("*"):
                del schema[k]
            else:
                remove_star_keys(v)
    elif isinstance(schema, (list, tuple)):
        for k in schema:
            remove_star_keys(k)


def use_star_keys(schema):
    if isinstance(schema, dict):
        for k, v in list(schema.items()):
            if k.startswith("*"):
                del schema[k]
                schema[k[1:]] = v
        for v in schema.values():
            use_star_keys(v)
    elif isinstance(schema, (list, tuple)):
        for k in schema:
            use_star_keys(k)



import re
from collections import namedtuple
from typing import Iterable, List

"""
A path is a sequence of names separated by '/' with an optional 
condition applied to the subtree.

The condition is a function taking current:PathCondition as a single parameter
which can be used for example as: 
current.a.b == "data" or current["a"]["b"] == "data" 

Any expression with the current should return an iterator or list of results (matched subtrees).
If any is truish, the condition is interpreted as matching.
"""

JSONPathRecord = namedtuple("JSONPathRecord", "path, condition, value")


class JSONPaths:
    def __init__(self):
        self.path_regex_list = []
        self.path_to_index = {}
        self.paths: List[List[JSONPathRecord]] = []

    def register(self, path, condition=None, value=None):
        """
        paths should be registered from longest to shortest

        :param path:        a path to register
        :param condition:   a condition applied to the node on the path
        :param value:       the value to return if path & condition match
        """

        if path not in self.path_to_index:
            self.path_to_index[path] = len(self.path_to_index)
            self.paths.append([])
            self.path_regex_list.append(re.compile("^" + path_to_regex(path) + "$"))
        path_locators = self.paths[self.path_to_index[path]]
        path_locators.append(
            JSONPathRecord(path=path, condition=condition, value=value)
        )

    def match(self, path=None, subtree=None, extra_data=None):
        """
        Matches a path and subtree against stored paths. Returns iterator of matched values

        :param path:        the path that should match
        :param subtree:     teh subtree against which to match stored conditions
        :return:            iterator of matched values
        """
        for idx, regex in enumerate(self.path_regex_list):
            match = regex.match(path)
            if not match:
                continue

            for rec in self.paths[idx]:
                if rec.condition:
                    condition_result = rec.condition(
                        PathCondition(subtree), **(extra_data or {})
                    )
                    if isinstance(condition_result, Iterable):
                        condition_result = any(condition_result)
                    if condition_result:
                        yield rec.value
                else:
                    yield rec.value


def path_to_regex(path):
    if isinstance(path, tuple):
        return "|".join(path_to_regex(x) for x in path)

    split_path = [x for x in re.split("(/)", path) if x]

    def fragment_to_regex(f):
        if f == "**":
            return ".+"
        return f.replace("*", "[^/]+")

    return "".join(fragment_to_regex(x) for x in split_path)


class PathCondition:
    def __init__(self, start=None, subtree_list=()):
        self._subtree_list = [*subtree_list]
        if start:
            self._subtree_list.append(start)

    def _apply(self, p, subtree_list):
        if subtree_list:
            if p == "*":
                for subtree in subtree_list:
                    itr = []
                    if isinstance(subtree, dict):
                        itr = subtree.values()
                    elif isinstance(subtree, list):
                        itr = subtree
                    yield from itr
            elif p == "**":
                first_level = list(self._apply("*", subtree_list))
                yield from first_level
                yield from self._apply("**", first_level)
            else:
                for subtree in subtree_list:
                    # match the path element and descend if found
                    if isinstance(subtree, dict):
                        if p in subtree:
                            yield subtree[p]
                    elif isinstance(subtree, list):
                        if isinstance(p, int) and 0 <= p < len(subtree):
                            yield subtree[p]

    def __getattr__(self, item):
        try:
            item = int(item)
        except:
            pass
        return PathCondition(subtree_list=list(self._apply(item, self._subtree_list)))

    def __getitem__(self, item):
        return PathCondition(subtree_list=list(self._apply(item, self._subtree_list)))

    def __eq__(self, other):
        if isinstance(other, PathCondition):
            # TODO: deep equals maybe
            return [
                subtree
                for subtree in self._subtree_list
                for o in other._subtree_list
                if subtree == o
            ]

        # TODO: deep equals maybe
        return [subtree for subtree in self._subtree_list if subtree == other]



import marshmallow as ma
from marshmallow import fields

from .utils import ExtendablePartSchema


class PropertySampleData(ExtendablePartSchema):
    skip = fields.Boolean(dump_default=False)
    faker = fields.String(required=False)
    params = fields.Raw(required=False)
    count = fields.Integer()


class ModelSampleConfiguration(ma.Schema):
    class RootSampleProps(ExtendablePartSchema):
        count = fields.Integer()

        class Meta:
            unknown = ma.RAISE

    sample = fields.Nested(RootSampleProps)



from oarepo_model_builder.utils.jinja import split_package_base_name

from ..builders import OutputBuilder
from ..outputs.cfg import CFGOutput


class InvenioCliSetupCfgBuilder(OutputBuilder):
    TYPE = "invenio_cli_setup_cfg"

    def finish(self):
        super().finish()

        output: CFGOutput = self.builder.get_output("cfg", "setup.cfg")

        cli_function = split_package_base_name(self.current_model.cli_function)

        output.add_entry_point(
            "flask.commands",
            self.current_model.flask_commands_setup_cfg,
            f"{'.'.join(cli_function[:-1])}:{cli_function[-1]}",
        )



import marshmallow as ma
from marshmallow import fields
from .model_validation import model_validator


class UIPropertyValidator(ma.Schema):
    marshmallow = fields.Nested(
        lambda: model_validator.validator_class("property-marshmallow")()
    )

    class Meta:
        unknown = ma.INCLUDE


class ObjectUIPropertyValidator(ma.Schema):
    marshmallow = fields.Nested(
        lambda: model_validator.validator_class("property-marshmallow-object")()
    )

    class Meta:
        unknown = ma.INCLUDE


class ObjectPropertyUISchema(ma.Schema):
    # x/[type object]
    ui = fields.Nested(lambda: model_validator.validator_class("object-property-ui")())



from oarepo_model_builder.utils.deepmerge import deepmerge


def test_merge_simple():
    assert deepmerge(1, 2, []) == 1


def test_merge_dict():
    assert deepmerge(dict(a=1), dict(b=2), []) == dict(a=1, b=2)
    assert deepmerge(dict(a=1), dict(a=2), []) == dict(a=1)

    assert deepmerge(dict(a=dict(a=1)), dict(a=dict(a=2)), []) == dict(a=dict(a=1))
    assert deepmerge(dict(a=dict(a=1)), dict(a=dict(b=2), c=3), []) == dict(
        a=dict(a=1, b=2), c=3
    )


def test_merge_list():
    assert deepmerge([1, 2], [3, 4], []) == [1, 2]
    assert deepmerge([1, 2], [3, 4, 5], []) == [1, 2, 5]


def test_merge_list_dict():
    assert deepmerge([dict(a=1)], [dict(b=1)], []) == [dict(a=1, b=1)]



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordServiceBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_service"
    class_config = "record-service-class"
    template = "record-service"



from pathlib import Path

import faker.config

from oarepo_model_builder.entrypoints import create_builder_from_entrypoints
from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.invenio.invenio_script_sample_data import (
    InvenioScriptSampleDataBuilder,
)
from oarepo_model_builder.schema import ModelSchema


def test_sample_builder_string():
    assert (
        build_sample_data(
            {
                "a": {
                    "type": "string",
                }
            }
        )
        == "a: test\n"
    )
    assert (
        build_sample_data(
            {
                "a": {
                    "type": "string",
                }
            },
            count=2,
        )
        == "a: test\n---\na: test\n"
    )


def test_sample_builder_integer():
    assert (
        build_sample_data(
            {
                "a": {
                    "type": "integer",
                }
            }
        )
        == "a: 1\n"
    )


def test_sample_builder_float():
    assert (
        build_sample_data(
            {
                "a": {
                    "type": "number",
                }
            }
        )
        == "a: 1.2\n"
    )


def test_sample_builder_date():
    assert (
        build_sample_data(
            {
                "a": {
                    "type": "date",
                }
            }
        )
        == "a: '2022-01-02'\n"
    )


def test_sample_builder_object():
    assert (
        build_sample_data(
            {
                "obj": {
                    "type": "object",
                    "properties": {
                        "a": {
                            "type": "string",
                        },
                        "b": {
                            "type": "string",
                        },
                    },
                }
            }
        ).strip()
        == """
obj:
  a: test
  b: test    
    """.strip()
    )


def test_sample_builder_simple_array():
    assert (
        build_sample_data(
            {
                "a": {
                    "type": "array",
                    "sample": {"count": 1},
                    "items": {"type": "string"},
                }
            }
        )
        == "a:\n- test\n"
    )

    # makes items unique, so expect one
    assert (
        build_sample_data(
            {
                "a": {
                    "type": "array",
                    "sample": {"count": 2},
                    "items": {"type": "string"},
                }
            }
        )
        == "a:\n- test\n"
    )


def test_sample_builder_complex_array():
    # sampler makes items in the array unique, so expect one
    assert (
        build_sample_data(
            {
                "a": {
                    "type": "array",
                    "sample": {"count": 2},
                    "items": {
                        "type": "object",
                        "properties": {
                            "b": {
                                "type": "string",
                            },
                            "c": {
                                "type": "string",
                            },
                        },
                    },
                }
            }
        ).strip()
        == """
a:
- b: test
  c: test
    """.strip()
    )


def build_sample_data(model, count=1):
    faker.config.PROVIDERS.clear()
    faker.config.PROVIDERS.append("tests.faker_constant")
    builder = create_builder_from_entrypoints()
    builder.filesystem = InMemoryFileSystem()
    builder.output_dir = Path.cwd()
    sample_builder = InvenioScriptSampleDataBuilder(
        builder=builder, property_preprocessors=[]
    )
    schema = ModelSchema(
        "test.json",
        {
            "model": {"script-import-sample-data": "test.yaml", "properties": model},
            "settings": {},
            "sample": {"count": count},
        },
    )
    sample_builder.build(schema)
    builder._save_outputs()
    sample_data = builder.filesystem.read("test.yaml")
    return sample_data



import json
import shutil
import sys
from functools import partial
from pathlib import Path

import click
import json5
import libcst as cst
import yaml

from oarepo_model_builder.utils.cst import PythonContext, merge
from oarepo_model_builder.utils.deepmerge import deepmerge


@click.command()
@click.argument("source", type=click.Path(file_okay=True, dir_okay=True, exists=True))
@click.argument(
    "destination", type=click.Path(file_okay=True, dir_okay=True, exists=True)
)
@click.option(
    "--result",
    type=click.Path(file_okay=True, dir_okay=True),
    help="Do not overwrite destination, write to this path instead",
)
@click.option(
    "--destination-first/--source-first",
    help="Take destination first and merge source only if it is not in destination. Normally merges destination into source and replaces destination with the result",
)
@click.option(
    "--overwrite/--no-overwrite",
    help="Do not perform merging but overwrite destination files with source files",
)
def merger(source, destination, result, destination_first, overwrite):
    source = Path(source)
    destination = Path(destination)

    if source.is_file():
        if not destination.is_file():
            click.echo(
                "If source is a file, destination must be a file as well", err=True
            )
            sys.exit(1)
        merge_file(
            source, destination, result or destination, destination_first, overwrite
        )
    else:
        for fn in source.glob("**/*"):
            if not fn.is_file():
                continue
            relative_fn = fn.relative_to(source)
            merge_file(
                fn,
                destination.joinpath(relative_fn),
                result.joinpath(relative_fn)
                if result
                else destination.joinpath(relative_fn),
                destination_first,
                overwrite,
            )


def merge_file(
    source: Path,
    destination: Path,
    result: Path,
    destination_first: bool,
    overwrite: bool,
):
    source_suffix = source.suffix[1:]
    destination_suffix = destination.suffix[1:]
    if source_suffix != destination_suffix:
        raise AttributeError(
            f"Suffixes of source and destination files must match, have {source}, {destination}"
        )
    result.parent.mkdir(parents=True, exist_ok=True)
    if not destination.exists() or overwrite:
        if source.is_file():
            shutil.copy(source, result)
        else:
            shutil.copytree(source, result, dirs_exist_ok=True)
        return
    if source_suffix == "yaml":
        merge_json(
            source,
            destination,
            result,
            destination_first,
            yaml.safe_load,
            yaml.safe_dump,
        )
    elif source_suffix == "json":
        merge_json(
            source,
            destination,
            result,
            destination_first,
            json.load,
            partial(json.dump, indent=4, ensure_ascii=False),
        )
    elif source_suffix == "json5":
        merge_json(
            source,
            destination,
            result,
            destination_first,
            json5.load,
            partial(json5.dump, indent=4, ensure_ascii=False),
        )
    elif source_suffix == "py":
        merge_python(source, destination, result, destination_first)
    else:
        raise NotImplementedError(
            f"Merging file type {source_suffix} (file {source}) is not implemented yet"
        )


def merge_json(
    source, destination, result: Path, destination_first, load_func, dump_func
):
    with open(source) as f:
        source_data = load_func(f)
    with open(destination) as f:
        destination_data = load_func(f)
    if destination_first:
        deepmerge(destination_data, source_data)
    else:
        deepmerge(source_data, destination_data)
        destination_data = source_data
    with open(result, "w") as f:
        dump_func(destination_data, f)


def merge_python(source: Path, destination: Path, result, destination_first: bool):
    source_cst = cst.parse_module(source.read_text())
    destination_cst = cst.parse_module(destination.read_text())
    if destination_first:
        result_cst = merge(
            PythonContext(source_cst),
            destination_cst,
            source_cst,
        )
    else:
        result_cst = merge(
            PythonContext(destination_cst),
            source_cst,
            destination_cst,
        )
    # write the merged files out
    with open(result, "w") as f:
        f.write(result_cst.code)

    # code style
    import subprocess

    import isort

    config = isort.settings.Config(verbose=False, quiet=True)
    isort.file(result, config=config)
    subprocess.call(["black", "-q", "--preview", str(result)])


if __name__ == "__main__":
    merger()
